# -*- python -*-
# $RCSfile: analysisdomain.py,v $
# $Revision: 1.32 $
# $Author: vrc $
# $Date: 2007-09-26 22:02:33 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Domain objects.  These are part of the post-solution mesh analysis
# process.  A "domain" is a user-specified portion of a mesh or that
# mesh's microstructure, specified in terms of mesh or microstructure
# objects, like pixels, elements, cross-sections, etc.

# One cannot actually evalute an output on a domain -- it's necessary
# to sample the domain first, and the sample objects are the
# numerically savvy components.  There can be more than one sampling
# of a domain.

from ooflib.SWIG.common import ooferror
from ooflib.common import microstructure
from ooflib.common import primitives
from ooflib.common import registeredclass
from ooflib.common.IO import parameter
from ooflib.common.IO import pixelgroupparam
from ooflib.common.IO import selectionplaceholder
from ooflib.common.IO import whoville
from ooflib.common.IO import xmlmenudump
from ooflib.engine import skeletoncontext
from ooflib.engine.IO import meshcsparams
from ooflib.engine.IO import skeletongroupparams
from ooflib.common import parallel_enable
import ooflib.engine.mesh

import sys

class Domain(registeredclass.RegisteredClass):
    registry = []
    def set_mesh(self, mesh):
        self.mesh = mesh
        self.mesh_obj = ooflib.engine.mesh.meshes[mesh]
        self.femesh = self.mesh_obj.getObject()
        self.skeleton = self.mesh_obj.skeleton
##        self.skelcontext = skeletoncontext.skeletonContexts[
##            self.skeleton.name()]
        self.skelcontext = skeletoncontext.skeletonContexts[
            skeletoncontext.extractSkeletonPath(mesh)]
        self.ms = self.mesh_obj.getMicrostructure()
        self.mscontext = microstructure.getMSContextFromMS(self.ms)
    def read_lock(self):
        self.mesh_obj.begin_reading()
        self.skelcontext.begin_reading()
        self.mscontext.begin_reading()
    def read_release(self):
        self.mscontext.end_reading()
        self.skelcontext.end_reading()
        self.mesh_obj.end_reading()

    tip = "A specified region of &mesh; or its &micro; for sampling."
    discussion = xmlmenudump.loadFile('DISCUSSIONS/engine/reg/domain.xml')


# Each subclass implements some subset of retrieval functions, which
# return information about the domain, in the form of a bounding-box
# rectangle, or a list of objects (elements, points).  The caller is
# expected to be something that will then construct sample objects, or
# wrap the returned objects in sample objects.

# Retrieval functions are things like get_bounds, get_pixels,
# get_endpoints.  Not all retrieval functions exist for all domains,
# and some retrieval functions might be expensive.  They should return
# None or an empty list if the domain is empty.

# Units, where they make sense, are physical units on the
# microstructure.

# Domains need to indicate what kinds of objects make sense for
# sampling, using the constants defined here.
CS_POINT = "cs_point_sample" # get_endpoints and get_elemental_segments.
POINT =    "point_sample"   # Promises existence of get_bounds, for the bbox.
PIXEL =    "pixel_sample"   # Promises existence of get_pixels.
ELEMENT =  "element_sample"  # Promises existence of get_elements.


# An object for operating on the entire mesh.
class EntireMesh(Domain):
    # Returns a bounding rectangle.
    def get_bounds(self):
        if parallel_enable.enabled():
            return self.skeleton.localbounds
        else:
            return primitives.Rectangle(primitives.Point(0.0,0.0),
                                        self.ms.size())
    
    # Returns a list of iPoint pixel indices.
    def get_pixels(self):
        return self.ms.coords()

    # Returns a list of all the elements.  This maybe should return the
    # iterator, to save memory?  The number of elements could be large.
    def get_elements(self):
        el_list = []
        iter = self.femesh.element_iterator()
        while not iter.end():
            el_list.append(iter.element())
            iter.next()
        return el_list
        

registeredclass.Registration(
    'Entire Mesh',
    Domain,
    EntireMesh,
    0,
    params=[],
    sample_types = [POINT, PIXEL, ELEMENT],
    tip='Use the entire Mesh as a post-processing domain.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/entire_mesh.xml'))

# Pixel selection or group -- these actually live on
# the host microstructure, so you have to go there to get 'em.
class PixelGroup(Domain):
    def __init__(self, pixels):
        self.pixels = pixels
    def _pixels_from_group(self):
        if self.pixels == selectionplaceholder.selectionPlaceHolder:
            return  self.ms.pixelselection.getSelection()
        elif self.pixels == selectionplaceholder.allPlaceHolder:
            return self.ms.coords()
        else:
            group = self.ms.findGroup(self.pixels)
            if group:
                return group.members()
    def get_pixels(self):
        return self._pixels_from_group()
    def get_bounds(self):
        ms_size = self.ms.size()
        pixels = self._pixels_from_group()
        if pixels:
            maxx=0.0
            maxy=0.0
            minx=ms_size[0]
            miny=ms_size[1]
            pixel_size = self.ms.sizeOfPixels()

            if parallel_enable.enabled():
                boundsxmin=self.skeleton.localbounds.xmin()
                boundsxmax=self.skeleton.localbounds.xmax()
                boundsymin=self.skeleton.localbounds.ymin()
                boundsymax=self.skeleton.localbounds.ymax()
                for p in pixels:
                    low_x = p[0]*pixel_size[0]
                    hi_x = low_x + pixel_size[0]
                    low_y = p[1]*pixel_size[1]
                    hi_y = low_y + pixel_size[1]
                    if low_x >= boundsxmin:
                        minx = min(minx, low_x)
                    if low_y >= boundsymin:
                        miny = min(miny, low_y)
                    if hi_x <= boundsxmax:
                        maxx = max(maxx, hi_x)
                    if hi_y <= boundsymax:
                        maxy = max(maxy, hi_y)
            else:
                for p in pixels:
                    low_x = p[0]*pixel_size[0]
                    hi_x = low_x + pixel_size[0]
                    low_y = p[1]*pixel_size[1]
                    hi_y = low_y + pixel_size[1]
                    minx = min(minx, low_x)
                    miny = min(miny, low_y)
                    maxx = max(maxx, hi_x)
                    maxy = max(maxy, hi_y)

            return primitives.Rectangle(primitives.Point(minx, miny),
                                        primitives.Point(maxx, maxy))
    
registeredclass.Registration(
    'Pixel Group',
    Domain,
    PixelGroup,
    10,
    params=[pixelgroupparam.PixelAggregateParameter('pixels',
                                                    tip=parameter.emptyTipString)],
    sample_types = [POINT, PIXEL],
    tip="Use a pixel group as a post-processing domain.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/pixelgroup_domain.xml'))

# Cross-section -- currently, all cross-sections are straight, so the
# endpoints provide a complete description.  Also, the elemental
# segments are straight.  This is named "CrossSectionDomain" to
# distinguish it from the CrossSection class which represents actual
# cross-sections in the mesh.
class CrossSectionDomain(Domain):
    def __init__(self, cross_section):
        self.cross_section=cross_section
        
    def get_endpoints(self):
        cs_obj = self.mesh_obj.getCrossSection(self.cross_section)

        # Clamp the end-points to be interior to the mesh.
        bounds = self.mesh_obj.size()
        
        boundvecs = [ primitives.Segment(primitives.Point(0.0,0.0),
                                         primitives.Point(bounds.x,0.0)),
                      primitives.Segment(primitives.Point(bounds.x,0.0),
                                         primitives.Point(bounds.x,bounds.y)),
                      primitives.Segment(primitives.Point(bounds.x,bounds.y),
                                         primitives.Point(0.0,bounds.y)),
                      primitives.Segment(primitives.Point(0.0,bounds.y),
                                         primitives.Point(0.0,0.0)) ]
        for s in boundvecs:
            if (s.end()-s.start()).cross(cs_obj.start-s.start()) < 0 and \
               (s.end()-s.start()).cross(cs_obj.end-s.start()) < 0:
                raise ooferror.ErrUserError(
                    "Segment with end-points %s, %s is entirely exterior to mesh %s." % (cs_obj.start, cs_obj.end, self.mesh_obj.name()) )
            
        real_start = cs_obj.start
        real_end = cs_obj.end

        # Vertical and horizontal cases don't occur -- if the
        # cross-section is vertical and out of bounds in x, then it's
        # entirely out of bounds, and we didn't get here.  Likewise,
        # it can't be horizontal and out of bounds in y.

        # Clip both points.
        if real_start.x < 0:
            alpha = real_start.x/(real_start.x-real_end.x)
            real_start.y = real_start.y + alpha*(real_end.y-real_start.y)
            real_start.x = 0 # Fix roundoff.
            
        if real_start.x > bounds.x:
            alpha = (real_start.x-bounds.x)/(real_start.x-real_end.x)
            real_start.y = real_start.y + alpha*(real_end.y-real_start.y)
            real_start.x = bounds.x
        
        if real_start.y < 0:
            alpha = (real_start.y)/(real_start.y-real_end.y)
            real_start.x = real_start.x + alpha*(real_end.x-real_start.x)
            real_start.y = 0

        if real_start.y > bounds.y:
            alpha = (real_start.y-bounds.y)/(real_start.y-real_end.y)
            real_start.x = real_start.x + alpha*(real_end.x-real_start.x)
            real_start.y = bounds.y

            
        if real_end.x < 0:
            alpha = (real_start.x)/(real_start.x-real_end.x)
            real_end.y = real_start.y + alpha*(real_end.y-real_start.y)
            real_end.x = 0

        if real_end.x > bounds.x:
            alpha = (real_start.x-bounds.x)/(real_start.x-real_end.x)
            real_end.y = real_start.y + alpha*(real_end.y-real_start.y)
            real_end.x = bounds.x
        
        if real_end.y < 0:
            alpha = (real_start.y)/(real_start.y-real_end.y)
            real_end.x = real_start.x + alpha*(real_end.x-real_start.x)
            real_end.y = 0

        if real_end.y > bounds.y:
            alpha = (real_start.y-bounds.y)/(real_start.y-real_end.y)
            real_end.x = real_start.x + alpha*(real_end.x-real_start.x)
            real_end.y = bounds.y

        return (real_start, real_end)
        # return (cs_obj.start, cs_obj.end)

    # Builds (and returns) a list of tuples, each consisting of a
    # primitives.Segment object and a mesh element.  These occur in
    # order from start to end, which map exactly once on to the cross
    # section, and whose boundaries correspond to intersections of the
    # cross-section with element boundaries.
    def get_elemental_segments(self):
        cs_obj = self.mesh_obj.getCrossSection(self.cross_section)

        (real_start, real_end) = self.get_endpoints()
        
        local_segment = primitives.Segment(real_start, real_end)
        skeleton = self.skeleton

        start_el = skeleton.enclosingElement(real_start)

        result = []

        # "last" means "most recent", and is distinct from "final".
        last_pt = cs_obj.start
        last_el = start_el

        try:
            # HACK: Because of round-off, the first call to this
            # function can see two intersections, but the first one is
            # spurious.  If this happens, re-run it, but tell it about
            # the spurious intersection.
            (isec, new_el) = \
                   skeleton.get_intersection_and_next_element(
                local_segment, last_el, None)
        except ooferror.ErrPyProgrammingError, e:
            if e.message()=="Segment exits element multiple times.":
                (isec, new_el) = \
                       skeleton.get_intersection_and_next_element(
                    local_segment, last_el, real_start)
            else:
                raise 
            

        # If the *first* call yields no intersections, then the
        # whole cross-section lies within one element.  Return
        # a single segment made up of the whole cross-section.
        if not isec:
            result.append( (primitives.Segment(last_pt, real_end),
                            self.femesh.getElement(last_el.meshindex))  )
            return result

        # Otherwise, keep iterating until you get to the end.
        while new_el:
            result.append( (primitives.Segment(last_pt, isec),
                            self.femesh.getElement(last_el.meshindex))  )
            last_pt = isec
            last_el = new_el
            (isec, new_el) = \
                   skeleton.get_intersection_and_next_element(
                local_segment, last_el, last_pt)

        # Bolt on the final one.
        result.append( (primitives.Segment(last_pt, real_end),
                        self.femesh.getElement(last_el.meshindex))  )

        return result


registeredclass.Registration(
    'Cross Section',
    Domain,
    CrossSectionDomain,
    20,
    params=[meshcsparams.MeshCrossSectionParameter('cross_section',
                                      tip='The cross section to sample.')],
    sample_types = [CS_POINT],
    tip='Use a Mesh cross section as the post-processing domain.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/xsection_domain.xml')
    )


class ElementGroup(Domain):
    def __init__(self, elements):
        self.elements = elements
    def _elements_from_aggregate(self):
        skeleton = self.skeleton
        skel_els = []
        if self.elements == selectionplaceholder.selectionPlaceHolder:
            skel_els = self.skelcontext.elementselection.retrieveFromSkeleton(
                skeleton)
        else:
            skel_els = self.skelcontext.elementgroups.get_groupFromSkeleton(
                self.elements, skeleton)
        return [self.femesh.getElement(s.meshindex) for s in skel_els]

    def get_bounds(self):
        first = 1
        for e in self._elements_from_aggregate():
            for n in e.cornernode_iterator():
                pos = n.position()
                if first:
                    xmin = pos[0]
                    xmax = pos[0]
                    ymin = pos[1]
                    ymax = pos[1]
                    first = None
                else:
                    xmin = min(pos[0], xmin)
                    xmax = max(pos[0], xmax)
                    ymin = min(pos[1], ymin)
                    ymax = max(pos[1], ymax)
        if not first:
            return primitives.Rectangle(primitives.Point(xmin, ymin),
                                        primitives.Point(xmax, ymax) )
                
    def get_elements(self):
        return self._elements_from_aggregate()

registeredclass.Registration(
    'Element Group',
    Domain,
    ElementGroup,
    30,
    params=[skeletongroupparams.ElementAggregateParameter('elements',
                                                          tip='Elements to sample.')],
    sample_types=[POINT, ELEMENT],
    tip='Use an element group as the post-processing domain.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/element_group_domain.xml')
    )

